indexing

	description:

		"'erl_g' command line tool"

	copyright: "Copyright (c) 2004-2005, Andreas Leitner and others"
	license: "Eiffel Forum License v2 (see forum.txt)"
	date: "$Date$"
	revision: "$Revision$"

class ERL_G

inherit

	ERL_G_COMMAND_LINE_PARSER
	KL_SHARED_STRING_EQUALITY_TESTER
	KL_SHARED_EXECUTION_ENVIRONMENT

creation

	execute

feature -- Execution

	execute is
			-- Start 'erl_g' execution.
		local
			a_file: KL_TEXT_INPUT_FILE
			a_lace_parser: ET_LACE_PARSER
			a_lace_error_handler: ET_LACE_ERROR_HANDLER
			an_xace_parser: ET_XACE_UNIVERSE_PARSER
			an_xace_error_handler: ET_XACE_DEFAULT_ERROR_HANDLER
			an_xace_variables: DS_HASH_TABLE [STRING, STRING]
			a_universe: ET_UNIVERSE
			nb: INTEGER
		do
			make
			Arguments.set_program_name ("erl_g")
			output_dirname := "reflection_library"
			create error_handler.make
			process_arguments

			create a_file.make (ace_filename)
			a_file.open_read
			if a_file.is_open_read then
				nb := ace_filename.count
				if nb > 5 and then ace_filename.substring (nb - 4, nb).is_equal (".xace") then
					create an_xace_error_handler.make_standard
					create an_xace_variables.make_map (100)
					an_xace_variables.set_key_equality_tester (string_equality_tester)
					if defined_variables /= Void then
						parse_defined_variables (defined_variables, an_xace_variables)					end
					create an_xace_parser.make_with_variables (an_xace_variables, an_xace_error_handler)
					an_xace_parser.parse_file (a_file)
					a_file.close
					if not an_xace_error_handler.has_error then
						a_universe := an_xace_parser.last_universe
					end
				else
					create a_lace_error_handler.make_standard
					create a_lace_parser.make (a_lace_error_handler)
					a_lace_parser.parse_file (a_file)
					a_file.close
					if not a_lace_parser.syntax_error then
						a_universe := a_lace_parser.last_universe
					end
				end
				if a_universe /= Void then
					process_universe (a_universe)
					if a_universe.error_handler.has_error then
						Exceptions.die (2)
					end
				else
					Exceptions.die (3)
				end
			else
					error_handler.report_cannot_read_error (ace_filename)
					Exceptions.die (1)
			end
		end

feature -- Parsing

	parse_defined_variables (an_input: STRING; a_table: DS_HASH_TABLE [STRING, STRING]) is
			-- Parse variable definitions in `an_input' and set the variables in `a_table' accordingly.
		require
			an_input_not_void: an_input /= Void
			a_table_not_void: a_table /= Void
		local
			a_splitter: ST_SPLITTER
			a_cursor: DS_LIST_CURSOR [STRING]
			a_definition: STRING
			an_index: INTEGER
			gobo_eiffel: STRING
		do
			gobo_eiffel := Execution_environment.variable_value ("GOBO_EIFFEL")
			if gobo_eiffel /= Void then
				a_table.force_last (gobo_eiffel, "GOBO_EIFFEL")
			end
			a_table.force_last ("True", "PROCESSOR_IS_ERL_G")
			create a_splitter.make
			a_cursor := a_splitter.split (defined_variables).new_cursor
			from a_cursor.start until a_cursor.after loop
				a_definition := a_cursor.item
				if a_definition.count > 0 then
					an_index := a_definition.index_of ('=', 1)
					if an_index = 0 then
						a_table.force_last ("", a_definition)
					elseif an_index = a_definition.count then
						a_table.force_last ("", a_definition.substring (1, an_index - 1))
					elseif an_index /= 1 then
						a_table.force_last (a_definition.substring (an_index + 1 ,a_definition.count), a_definition.substring (1, an_index - 1))
					end
				end
				a_cursor.forth
			end
		end

feature {NONE} -- Processing

	process_universe (a_universe: ET_UNIVERSE) is
			-- Process `a_universe'.
		require
			a_universe_not_void: a_universe /= Void
		local
			an_ast_factory: ET_DECORATED_AST_FACTORY
		do
			create an_ast_factory.make
			an_ast_factory.set_keep_all_breaks (True)
			a_universe.set_ast_factory (an_ast_factory)

			a_universe.error_handler.set_ise
			a_universe.set_use_assign_keyword (True)
			a_universe.set_use_attribute_keyword (False)
			a_universe.set_use_convert_keyword (True)
			a_universe.set_use_create_keyword (True)
			a_universe.set_use_recast_keyword (True)
			a_universe.set_use_reference_keyword (False)
			if void_feature then
				a_universe.set_use_void_keyword (False)
			else
				a_universe.set_use_void_keyword (True)
			end
			a_universe.compile_all (False)
			generate_reflection_classes (a_universe)
		end

	generate_reflection_classes (a_universe: ET_UNIVERSE) is
			-- Generate reflection classes for universe `a_universe'.
		require
			a_universe_not_empty: a_universe /= Void
		local
			cs: DS_LINEAR_CURSOR [STRING]
			l_class_type: ET_CLASS_TYPE
			reflection_generator: ERL_G_GENERATOR
		do
			create reflection_generator.make (a_universe)
			from
				cs := type_names.new_cursor
				cs.start
			until
				cs.off
			loop
				l_class_type := class_type (cs.item, a_universe)
				if l_class_type = Void then
					error_handler.report_invalid_or_unknown_type_error (cs.item)
					Exceptions.die (1)
				else
					reflection_generator.mark_base_type_reflectable (l_class_type)
				end
				cs.forth
			end
			reflection_generator.generate_all (output_dirname)
			if reflection_generator.has_fatal_error then
				error_handler.report_fatal_generation_error
			end
		end

feature {NONE} -- Parsing class types

	class_type (a_name: STRING; a_universe: ET_UNIVERSE): ET_CLASS_TYPE is
			-- Class type of `a_name' or `Void' if it doesn't exist
		require
			a_name_not_void: a_name /= Void
			a_universe_not_void: a_universe /= Void
		local
			i: INTEGER
		do
			i := parse_class_type (a_name, 1, a_universe)
			Result := last_class_type
			last_class_type := Void
		ensure
			valid_context: Result /= Void implies Result.is_valid_context
		end

	parse_class (str: STRING; a_position: INTEGER; a_universe: ET_UNIVERSE): INTEGER is
			-- Parse class in `str' starting at `a_position'.
			-- Make result available in `last_class', or Void
			-- if an error was found (report error to `output_file'.
			-- Return the new position in `str'.
		require
			str_not_void: str /= Void
			a_universe_not_void: a_universe /= Void
		local
			a_name: STRING
			i, nb: INTEGER
			c: CHARACTER
			an_identifier: ET_IDENTIFIER
			stop: BOOLEAN
		do
			last_class := Void
			from
				i := a_position
				nb := str.count
			until
				i > nb or
				stop
			loop
				c := str.item (i)
				inspect c
				when 'a'..'z', 'A'..'Z', '0'..'9', '_' then
					stop := True
				else
					i := i + 1
				end
			end
			from
				stop := False
				create a_name.make (10)
			until
				i > nb or
				stop
			loop
				c := str.item (i)
				inspect c
				when 'a'..'z', 'A'..'Z', '0'..'9', '_' then
					a_name.append_character (c)
					i := i + 1
				else
					stop := True
				end
			end
			Result := i
			if not a_name.is_empty then
				create an_identifier.make (a_name)
				last_class := a_universe.eiffel_class (an_identifier)
			else
				Result := nb + 2
			end
		ensure
			valid_position: Result > a_position
		end

	last_class: ET_CLASS
			-- Last class parsed by `parse_class';
			-- Void if an error was found when parsing

	parse_class_type (str: STRING; a_position: INTEGER; a_universe: ET_UNIVERSE): INTEGER is
			-- Parse class type in `str' starting at `a_position'.
			-- Make result available in `last_class_type', or Void
			-- if an error was found (report error to `output_file'.
			-- Return the new position in `str'.
		require
			str_not_void: str /= Void
			a_universe_not_void: a_universe /= Void
		local
			i, nb: INTEGER
			stop: BOOLEAN
			a_class: ET_CLASS
			an_actuals, tmp_actuals: ET_ACTUAL_PARAMETER_LIST
		do
			last_class_type := Void
			Result := parse_class (str, a_position, a_universe)
			a_class := last_class
			if a_class /= Void then
				if a_class.is_generic then
					Result := parse_open_bracket (str, Result)
					if Result > str.count + 1 then
						stop := True
					end
					nb := a_class.formal_parameters.count
					create tmp_actuals.make_with_capacity (nb)
					from
						i := 1
					until
						i > nb or stop
					loop
						Result := parse_class_type (str, Result, a_universe)
						if last_class_type /= Void then
							tmp_actuals.put_first (last_class_type)
							if i /= nb then
								Result := parse_comma (str, Result)
								if Result > str.count + 1 then
									stop := True
								end
							end
						else
							stop := True
							Result := str.count + 2
						end
						i := i + 1
					end
					last_class_type := Void
					if not stop then
						Result := parse_close_bracket (str, Result)
						if Result <= str.count + 1 then
							create an_actuals.make_with_capacity (nb)
							from i := nb until i < 1 loop
								an_actuals.put_first (tmp_actuals.item (i))
								i := i - 1
							end
							create {ET_GENERIC_CLASS_TYPE} last_class_type.make (Void, a_class.name, an_actuals, a_class)
						end
					end
				else
					last_class_type := a_class
				end
			end
		ensure
			valid_position: Result > a_position
			valid_context: last_class_type /= Void implies last_class_type.is_valid_context
		end

	last_class_type: ET_CLASS_TYPE
			-- Last class type parsed by `parse_class_type';
			-- Void if an error was found when parsing

	parse_open_bracket (str: STRING; a_position: INTEGER): INTEGER is
			-- Parse '[' in `str' starting at `a_position'.
			-- Return the new position in `str', or
			-- 'str.count + 2' if no ']' was found.
		local
			i, nb: INTEGER
			found, error: BOOLEAN
			c: CHARACTER
		do
			from
				i := a_position
				nb := str.count
			until
				i > nb or
				found or error
			loop
				c := str.item (i)
				inspect c
				when '[' then
					i := i + 1
					found := True
				when ' ', '%T', '%R', '%N' then
						-- Skip spaces.
					i := i + 1
				else
					error := True
				end
			end
			if not found or error then
				Result := nb + 2
			else
				Result := i
			end
		ensure
			valid_position: Result > a_position
		end

	parse_comma (str: STRING; a_position: INTEGER): INTEGER is
			-- Parse ',' in `str' starting at `a_position'.
			-- Return the new position in `str', or
			-- 'str.count + 2' if no ']' was found.
		local
			i, nb: INTEGER
			found, error: BOOLEAN
			c: CHARACTER
		do
			from
				i := a_position
				nb := str.count
			until
				i > nb or
				found or error
			loop
				c := str.item (i)
				inspect c
				when ',' then
					i := i + 1
					found := True
				when ' ', '%T', '%R', '%N' then
						-- Skip spaces.
					i := i + 1
				else
					error := True
				end
			end
			if not found or error then
				Result := nb + 2
			else
				Result := i
			end
		ensure
			valid_position: Result > a_position
		end

	parse_close_bracket (str: STRING; a_position: INTEGER): INTEGER is
			-- Parse ']' in `str' starting at `a_position'.
			-- Return the new position in `str', or
			-- 'str.count + 2' if no ']' was found.
		local
			i, nb: INTEGER
			found, error: BOOLEAN
			c: CHARACTER
		do
			from
				i := a_position
				nb := str.count
			until
				i > nb or
				found or error
			loop
				c := str.item (i)
				inspect c
				when ']' then
					i := i + 1
					found := True
				when ' ', '%T', '%R', '%N' then
						-- Skip spaces.
					i := i + 1
				else
					error := True
				end
			end
			if not found or error then
				Result := nb + 2
			elseif found then
				from
					found := False
				until
					i > nb or
					found
				loop
					c := str.item (i)
					inspect c
					when ' ', '%T', '%R', '%N' then
							-- Skip spaces.
						i := i + 1
					else
						found := True
					end
				end
				Result := i
			else
				Result := i
			end
		ensure
			valid_position: Result > a_position
		end

end
